/* avl's with state packing */

/* define SETJMP if state comparison is expensive */
// #define SETJMP

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#ifdef SETJMP
#include <setjmp.h>
#endif
#ifndef MINGW
#include <sys/types.h>
#include <sys/mman.h>
#include <unistd.h>
#ifdef __MACH__
#include <mach/vm_statistics.h>
#endif
#endif
#include "avl.h"



/* imported --------------------------------------------------------------------------------- */

extern size_t pack_value();
extern size_t unpack_value();
extern int    compare_packed_value();
extern size_t get_size();

#define compare_states compare_packed_value


/* declarations ----------------------------------------------------------------------------- */

// Memory management
#ifndef MINGW
static char* current_start = 0;
static size_t current_used = 0;
#ifdef __MACH__
// useful to profile memory usage on MacOS X
static const int vm_tag = VM_MAKE_TAG(240);
#else
static const int vm_tag = -1;
#endif
static const size_t block_size = 16777216;
#endif

static char* pack_buffer  = 0;   /* packing buffer (holds last compressed state followed by 0) */
static void* last_state   = 0;   /* unpacling buffer (holds last unpacked state) */
static key last_state_key = 0;   /* key of last unpacked state    */
static size_t value_size  = 0;


/* packing/unpacking routines --------------------------------------------------------------- */

/* packs state into pack_buffer by pack_value */
static char* pack_state (void* st)
{
  pack_value(st, pack_buffer);
  return pack_buffer;
}

/* creates a string from the state packed in pack_buffer (called if state found new) */
static char* copy_from_pack_buffer(char* place)
{
  const size_t size = get_size(pack_buffer);
  memcpy(place, pack_buffer, size);
  return place;
}

/* unpacks string into a state in last_state (returned as result) */
static char* unpack_state (char* str)
{
  if (last_state_key == (key)str) {
    return last_state;               /* cached; last state unpacked is left there */
  } else {
    last_state_key = (key)str;       /* miss: unpack str into last_state */
    unpack_value(last_state, str);
    return last_state;
  }
}

#ifdef MINGW
#define alloc(size) (malloc(size))
#else
static 
char*
alloc(size_t size)
{
  char* previous_start = 0;
  size_t previous_used  = current_used;
  
  // do we need to allocate more memory?
  if (block_size - current_used < size)
  {
    char* new_start = mmap(0, block_size, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PRIVATE, vm_tag, 0);
    if (new_start == MAP_FAILED)
    {
      return NULL;
    }

    madvise(new_start, block_size, MADV_RANDOM);

    // are the old block and the new block contiguous?
    if ((current_start + block_size) == new_start)
    {
      previous_start = current_start;
    }
    current_start = new_start;
    current_used = 0;
  }

  if (previous_start) // contiguous blocks
  {
    // don't forget to remove the size of what is stored in the old block
    current_used = size - (block_size - previous_used);
    return previous_start + previous_used;
  }
  else
  {
    previous_used = current_used; // beginning of the 'allocated' block
    current_used += size;
    return current_start + previous_used;
  }
}
#endif


/* avl storage routines --------------------------------------------------------------------- */

/* Warning! The allocated node is in fact an avl followed by the stored data
   Thus, the real size of an avl is sizeof(avl) + get_size(data)
*/

typedef struct __attribute__ ((__packed__)) avl {
  struct avl *left;
  struct avl *right;
  unsigned char height;
} avl;

static avl* storage = NULL;

#define Height(P) ((P==NULL) ? 0 : (P)->height)
#define Data(node)            ((char*)node + sizeof(avl))

#define Max(x,y) (((x)>(y)) ? (x) : (y))
#define Abs(x) (((x)<0) ? -(x) : (x))

static avl *SingleRotateWithLeft(avl *K2) {
  avl *K1;
  K1 = K2->left;
  K2->left = K1->right;
  K1->right = K2;
  K2->height = Max(Height(K2->left), Height(K2->right)) + 1;
  K1->height = Max(Height(K1->left), K2->height) + 1;
  return K1;
}

static avl *SingleRotateWithRight(avl *K1) {
  avl *K2;
  K2 = K1->right;
  K1->right = K2->left;
  K2->left = K1;
  K1->height = Max(Height(K1->left), Height(K1->right)) + 1;
  K2->height = Max(Height(K2->right), K1->height) + 1;
  return K2;
}

static avl *DoubleRotateWithLeft(avl *K3) {
  K3->left = SingleRotateWithRight(K3->left);
  return SingleRotateWithLeft(K3);
}

static avl *DoubleRotateWithRight (avl *K1) {
  K1->right = SingleRotateWithLeft(K1->right);
  return SingleRotateWithRight(K1);
}


static key result;

#ifdef SETJMP
static jmp_buf env;
#else
static int Lookup (void *e, avl *P) {
  avl *T=P;
  while (1) {
    if (T == NULL) {
      return 0;
    } else {
      switch (compare_states(e,Data(T))) {
      case -1 : 
	T=T->left;
	break;
      case  0 :
	result = (key)Data(T);
	return 1;
	break;
      case  1 :
	T=T->right;
	break;
      }
    }
  }
}
#endif

static avl *Insert (void *e, avl *T) {
  if (T == NULL) {
    T = (avl*)alloc(sizeof(avl) + get_size(pack_buffer));
    if (T == NULL)
      {fprintf(stderr,"tts exception: out of memory\n"); fflush(stderr); exit(1);}
    else {
      T->height = 1;
      T->left = T->right = NULL;
      copy_from_pack_buffer(((char*)T + sizeof(avl)));
      result=(key)Data(T);
    }
  } else {
    switch (compare_states(e,Data(T))) {
    case -1 : 
      T->left = Insert(e,T->left);
      if (Height(T->left) - Height(T->right) == 2)
	if (Height(T->left->right) < Height(T->left->left)) /* eqv. (compare_states(e,Data(T->left)) < 0) */
	  T = SingleRotateWithLeft(T);
	else
	  T = DoubleRotateWithLeft(T);
      T->height = Max(Height(T->left), Height(T->right)) + 1;
      break;
#ifdef SETJMP
    case  0 :
      result = (key)Data(T);
      longjmp(env,1);
      break;
#endif
    case  1 :
      T->right = Insert(e,T->right);
      if (Height(T->right) - Height(T->left) == 2)
	if (Height(T->right->right) > Height(T->right->left)) /* eqv. (compare_states(e,Data(T->right)) > 0) */
	  T = SingleRotateWithRight(T);
	else
	  T = DoubleRotateWithRight(T);
      T->height = Max(Height(T->left), Height(T->right)) + 1;
      break;
    }
  }
  return T;
}

/* SearchInsert (combines lookup and add) */
static key SearchInsert (void *e) {
#ifdef SETJMP
  if (!setjmp(env))
#else
  if (!Lookup(e,storage))
#endif
    storage = Insert (e,storage);
  return result;
}


/* for debug and statictics */

/* size of avl */
static int avlsize = 0;

/* returns height or -1 if avl property not met */
static int isAvl (avl *t) {
  int hl, hr;
  if (t==NULL) return 0;
  avlsize +=  sizeof(avl) + get_size((char*)t + sizeof(avl));
  if ((hl = isAvl (t->left)) < 0) return -1;
  if ((hr = isAvl (t->right)) < 0) return -1;
  if (Abs(hl - hr) > 1) return -1;  /* bad balance */
  if (t->height == 1 + Max(hl,hr)) return t->height;
  return -1;  /* bad recorder height */
}

/* checks isAvl and computes size */
static int check_avl () {
  int h;
  fflush(stderr);
  if ((h = isAvl(storage)) < 0) {
    fprintf(stderr,"\risavl = false\n");
    fprintf(stderr,"dataspace size = %d\n",avlsize);
  } else {
    fprintf(stderr,"\risavl = true, height=%d\n",h);
    fprintf(stderr,"dataspace size = %d\n",avlsize);
  }
  fflush(stderr);
  return h;
}



/* storage api routines --------------------------------------------------------------------- */

int init_storage(size_t size) {
  value_size = size;
#ifndef MINGW
  current_start = mmap(0, block_size, PROT_READ | PROT_WRITE, MAP_ANON | MAP_PRIVATE,  vm_tag, 0);
  if (current_start == MAP_FAILED) return 1;
  madvise(current_start, block_size, MADV_RANDOM);
#endif
  // +16 bytes : if sizeof(struct value) is too small (< (sizeof(int) + sizeof(int))/2)
  //                                                      ^ hash        ^ max size encoding
  // we size * 2 won't hold hash and size of a packed value
  pack_buffer = (char*)alloc(size*2 + 16);
  last_state = (void*)alloc(size);
  return 0;
}

int finalize_storage(int debug) {
  if (debug) check_avl ();
  return 0;
}

key store_state (void* state) {
  return SearchInsert(pack_state(state));
}

void* lookup_state (key k) {
  return unpack_state ((void *)k);
}


